//screensaver main drawing thread
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include "fractal.h"
#include "util0.h"
#include "bmp.h"
#include "dibsect.h"
#define PIXELPERSTEP 5

static BYTE staticHCST[8192];
BYTE*CalcHiColorShadowTable()
//in C++, this would be a static function
{int ii,shd;
 BYTE*pb;

  pb=staticHCST;
  for (ii=0;ii<32;ii++) { //RGB have 5-bit each, [0,31]
    //256 levels is overkill for high-color,
    //but just in case we switch to true-color later
    for (shd=0;shd<128;shd++) {//0-127: darken
      pb[0]=(ii*shd*2+128)>>8;
      pb++;
    }
    *pb=ii; pb++; //128: no change
    for (shd=129;shd<256;shd++) {//129-255: brighten
      pb[0]=ii+((31-ii)*(shd-128)*2+127)/254;
      pb++;
    }
  }
  return staticHCST;
}
void CastShadow(const RENDERINFO*pri,const RECT*prcri,
    const PIXELARRAY*ppa,int x0pa,int y0pa,const BYTE*pshd)
//pri->RENDERINFO
//prcri->rectangle inside RENDERINFO to draw in
//ppa->PIXELARRAY, this is the shadow mask
//x0pa,y0pa=offset of the starting pixel inside PIXELARRAY
//pshd->shadow lookup table
{int xsdraw,ysdraw,jj,ii;
 WORD*psurface,bred,bgrn,bblu;
 BYTE*pmask,shadow;

  xsdraw=prcri->right - prcri->left;
  ysdraw=prcri->bottom - prcri->top;

  for (jj=0;jj<ysdraw;jj++) {
    psurface=pri->pSurface + prcri->left +
             pri->BmpInfoHdr.biWidth*(prcri->top+jj);
    pmask=ppa->ppaArray + (y0pa+jj)*ppa->paX + x0pa;
    ii=xsdraw;
//the reason for this optimization is not speed, but CPU usage.
//because this is a screensaver, we want to programs running in
//background more CPU cycle
    _asm {
      push esi
      push edi
      mov edx,pmask
      mov edi,psurface
      mov esi,pshd
      xor ebx,ebx
      
actde:mov bl,[edx] //shadow=pmask[0]
      mov ax,[edi]
      mov bh,al
      and bh,0x1f
      mov cl,[esi][ebx] //blue
      shr ax,5
      mov bh,al
      and bh,0x1f
      mov ch,[esi][ebx] //green
      shr ax,5
      mov bh,al
      and bh,0x1f
      mov ah,[esi][ebx] //red
      mov al,ch
      shl al,3
      shl ax,2
      or al,cl
      mov [edi],ax
      inc edx
      add edi,2
      dec dword ptr ii
      jnz short actde
      pop edi
      pop esi
    }
/*equilavent C code
    for (ii=0;ii<xsdraw;ii++) {
      shadow=pmask[0];
      //FEDCBA9876543210
      //0RrrrrGggggBbbbb
      bred=(0x1f00&(psurface[0]>>2))|shadow;
      bgrn=(0x1f00&(psurface[0]<<3))|shadow;
      bblu=(0x1f00&(psurface[0]<<8))|shadow;
      psurface[0]=(pshd[bred]<<10)|(pshd[bgrn]<<5)|(pshd[bblu]);
      psurface++; pmask++;
    }
*/
  }
}


void DrawUninit(RENDERINFO*priscr,BYTE*prisave,PIXELARRAY*ppa)
{
  FreeMem(prisave);
  RenderObjDelete(priscr);
  PixelArrayDelete(ppa);
}

int DrawInit(RENDERINFO*priscr,BYTE**psave,PIXELARRAY*ppa,char*bmpfile)
/* initialization of drawing thread
1. load BMP into ppa
2. create a prisave with the same dimension as ppa
3. copy screen into priscr
4. init shadow/light table
*/
{int xsave,ysave;
 RGBQUAD pal[256];

  if (PixelArrayLoad256ClrBmp(ppa,pal,bmpfile)) return -1;
  PixelArray256ClrToGray(ppa,pal);

  if (0==RenderObjCreateCopyScr(priscr)) {
    xsave=GetSystemMetrics(SM_CXSCREEN);
    ysave=GetSystemMetrics(SM_CYSCREEN);
    if (xsave > ppa->paX) xsave=ppa->paX;
    if (ysave > ppa->paY) ysave=ppa->paY;
    *psave=GetMem((xsave*ysave)<<1);
    if (*psave) {
      CalcHiColorShadowTable();
      return 0;
    }
    RenderObjDelete(priscr);
  }
  PixelArrayDelete(ppa);
  return -1;
}
static char szCantLoadBmp[]={"Cannot load shdwss.bmp"};
DWORD WINAPI RenderProc(SSINFO*pInfo)
/*main screen saver painting procedure
WARNING: pInfo->hThread is in the context of main thread

main screen painting flow:
1. initialize
1a. if window not full screen set mapping mode
2. create starting point
3. calc OLDRECT based on starting point
4. create end point
5.  for each step
5a.   if (exit command) goto step90
5b.   calculate current step, DRAWRECT, etc
//no double-buffering here, slower but saves memory
5c.   save DRAWRECT 
5d.   draw current step in DRAWRECT
5e.   copy UnionRect(OLDRECT,DRAWRECT) to screen
5f.   restore DRAWRECT
5g.   OLDRECT=DRAWRECT
    next
6. starting point=end point
7. goto step 4
90. uninitialize
*/
{RENDERINFO riscr; //copy of screen
 BYTE*pSave; //for saving changed part
 PIXELARRAY pabmp; //our light/shadow mask
 HDC hdc;
 POINT pt0,pt1;
 RECT oldrc,newrc,scrrc;
 int WinXs,WinYs,curstep,stepscnt;
 char windir[300];
 
  GetWindowRect(GetDesktopWindow(),&scrrc);
  WinXs=scrrc.right-scrrc.left;
  WinYs=scrrc.bottom-scrrc.top;
  srand(GetTickCount());
  hdc=GetDC(pInfo->DrawWnd);

  //step 1:initialize
  GetModuleFileName(GetModuleHandle(NULL),windir,300);
  curstep=lstrlen(windir);
  while (curstep) { //search for last backslash
    if ('\\'==windir[curstep-1]) break;
    curstep--;
  }
  lstrcpy(windir+curstep,"shdwss.bmp");

  GetWindowRect(pInfo->DrawWnd,&newrc);
  if (DrawInit(&riscr,&pSave,&pabmp,windir)) { //if init failed
    PatBlt(hdc,newrc.left,newrc.top, newrc.right-newrc.left,
           newrc.bottom-newrc.top,BLACKNESS); //black out screen
    TextOut(hdc,0,0,szCantLoadBmp,lstrlen(szCantLoadBmp));
    //TextOut(hdc,0,0,windir,lstrlen(windir));
    ReleaseDC(pInfo->DrawWnd, hdc);
    return -1; //this exits the current thread
  }
  //step 1a. if window not full screen set mapping mode
  if (newrc.right-newrc.left!=WinXs ||
      newrc.bottom-newrc.top!=WinYs) {
    SetMapMode(hdc,MM_ANISOTROPIC);
    SetViewportExtEx(hdc,newrc.right-newrc.left,newrc.bottom-newrc.top,NULL);
    SetWindowExtEx(hdc,WinXs,WinYs,NULL);
    RenderObjToDc(&riscr,hdc,&scrrc,0,0);
  }
  
  //step 2: create starting point
  pt0.x=rand()%WinXs; pt0.y=rand()%WinYs;
  //step 3: calc OLDRECT based on starting point
  oldrc.left=pt0.x; oldrc.top=pt0.y;
  oldrc.right=pt0.x+1; oldrc.bottom=pt0.y+1;
  
rpbr30: //step 4: create end point
  pt1.x=rand()%WinXs; pt1.y=rand()%WinYs;

  curstep=abs(pt1.x-pt0.x); 
  stepscnt=abs(pt1.y-pt0.y);
  if (curstep<stepscnt) curstep=stepscnt;
  stepscnt=curstep/PIXELPERSTEP;

  for (curstep=0;curstep<stepscnt;curstep++) {
    int xnow,ynow; RECT drawrc;
    //the coordinate calculation code is not very efficient,
    //but it's not the bottleneck.

    //5a. if exit command goto step90;
    if (1 & (pInfo->dwCommand)) goto rpbr90;
    //5b. calculate current step, DRAWRECT, etc
    xnow=(pt1.x-pt0.x)*curstep/stepscnt+pt0.x-(pabmp.paX>>1);
    ynow=(pt1.y-pt0.y)*curstep/stepscnt+pt0.y-(pabmp.paY>>1);
    newrc.left=xnow; newrc.right=newrc.left+pabmp.paX;
    newrc.top=ynow; newrc.bottom=newrc.top+pabmp.paY;
    IntersectRect(&drawrc,&newrc,&scrrc);
    //5c. save DRAWRECT
    RenderObjGetRect(&riscr,&drawrc,pSave);
    //5d. draw current step in DRAWRECT
    CastShadow(&riscr,&drawrc,&pabmp, drawrc.left-newrc.left,
               drawrc.top-newrc.top, staticHCST);
    //5e. copy UnionRect(OLDRECT,DRAWRECT) to screen
    UnionRect(&newrc,&drawrc,&oldrc);
    RenderObjToDc(&riscr,hdc,&newrc,newrc.left,newrc.top);
    //5f. restore DRAWRECT
    RenderObjPutRect(&riscr,&drawrc,pSave);
    //5g. OLDRECT=DRAWRECT
    oldrc=drawrc;
    Sleep(50); //give background program more CPU time
  }
  pt0=pt1; //starting point=end point;
  goto rpbr30; //goto step 4

rpbr90:
  ReleaseDC(pInfo->DrawWnd,hdc);
  DrawUninit(&riscr,pSave,&pabmp);
  return 0;
}
